import json
import re
import warnings
from asyncio import sleep
from functools import wraps
from hashlib import md5
from pathlib import Path
from typing import Any, AsyncGenerator, Literal, Optional, Type, Union

import mercantile
import numpy as np
import ua_generator
from hishel import (
    AsyncBaseStorage,
    AsyncCacheTransport,
    AsyncFileStorage,
    AsyncRedisStorage,
    AsyncSQLiteStorage,
)
from httpx import (
    AsyncBaseTransport,
    AsyncClient,
    AsyncHTTPTransport,
    ConnectError,
    HTTPError,
    ProxyError,
    RemoteProtocolError,
    Response,
    TimeoutException,
)
from httpx._types import ProxyTypes, QueryParamTypes
from shapely import MultiPolygon, Point, Polygon, box, to_geojson

try:
    import anysqlite
except ImportError:
    anysqlite = None  # type: ignore
try:
    import redis.asyncio as redis
except ImportError:
    redis = None  # type: ignore

import pynspd.errors as err
from pynspd.client import (
    NSPD_CACHE_CONTROLLER,
    SSL_CONTEXT,
    BaseNspdClient,
)
from pynspd.logger import logger
from pynspd.map_types.enums import TabTitle, ThemeId
from pynspd.schemas import Layer36048Feature, Layer36049Feature, NspdFeature
from pynspd.schemas.feature import Feat
from pynspd.schemas.responses import (
    NspdTabGroupResponse,
    NspdTabResponse,
    SearchResponse,
)


def retry_on_http_error(func):
    """Декоратор для повторения запроса при ошибках запроса"""

    @wraps(func)
    async def wrapper(self: "AsyncNspd", *args, **kwargs):
        attempt = -1
        last_error: Optional[Exception] = None
        while True:
            attempt += 1
            logger_suffix = f'Wrapped method "{func.__name__}", retry {attempt} -'
            if attempt > self._retries:
                logger.debug("%s run out attempts", logger_suffix)
                if last_error is None:
                    last_error = Exception(
                        f"Too many retries ({attempt}/{self._retries})"
                    )
                raise last_error
            try:
                logger.debug("%s start request", logger_suffix)
                return await func(self, *args, **kwargs)
            except HTTPError as e:
                last_error = e
                if isinstance(e, RemoteProtocolError):
                    logger.debug("%s remote protocol error", logger_suffix)
                elif isinstance(e, ConnectError):
                    logger.debug("%s connection error", logger_suffix)
                elif isinstance(e, TimeoutException):
                    logger.debug("%s timeout", logger_suffix)
                    await sleep(1)
                else:
                    logger.exception("%s unexpected exception", logger_suffix)
                    raise e
            except err.PynspdServerError as e:
                logger.debug("%s unknown server error", logger_suffix)
                last_error = e
            except err.BlockedIP as e:
                logger.debug("%s blocked IP, retrying", logger_suffix)
                last_error = e
                if not self._retry_on_blocked_ip:
                    raise e
            except err.TooManyRequests as e:
                logger.debug("%s too many requests", logger_suffix)
                last_error = e
                await sleep(1)

    return wrapper


class AsyncNspd(BaseNspdClient):
    """Асинхронный клиент для НСПД

    Example:
    ```python
    async with pynspd.AsyncNspd() as nspd:
        feat = await nspd.find_zu("77:05:0001005:19")
    ```

    Args:
        client_timeout:
            Время ожидания ответа.
            Если не установлен - есть вероятность бесконечного ожидания. По умолчанию `None`.
        client_retries:
            Количество попыток при неудачном запросе
            (таймаут, неожиданный обрыв соединения, 5хх ошибки). По умолчанию `10`.
        client_retry_on_blocked_ip:
            При получении ошибки 403 (доступ заблокирован для вашего IP),
            продолжать попытки запроса до исчерпания retries.
            Рекомендуется использовать только c ротируемыми прокси. По умолчанию `False`.
        client_proxy:
            Адрес для проксирования запросов. По умолчанию `None`.
        client_dns_resolve:
            Использовать в запросах IP адрес НСПД вместо доменного имени.
            Рекомендуется включить, если используемый прокси
            не может сам разрешать доменные имена.
            По умолчанию `False`.
        cache_folder_path:
            Путь до папки для кэша запросов.. По умолчанию `None`.
        cache_sqlite_url:
            Строка подключения для sqlite-хранилища кэша. По умолчанию `None`.
        cache_redis_url:
            Строка подключения для redis-хранилища кэша. По умолчанию `None`.
        cache_ttl:
            Количество времени (в секундах) сколько будет храниться кэш. По умолчанию `None`.
        cache_storage:
            Ручная настройка объекта хранилища кэша (см. https://hishel.com/advanced/storages/). По умолчанию `None`.
        trust_env:
            Использовать переменные окружения для инициализации.
            По умолчанию `True`.
    """

    def __init__(
        self,
        *,
        client_timeout: Optional[int] = None,
        client_retries: Optional[int] = None,
        client_retry_on_blocked_ip: Optional[bool] = None,
        client_proxy: Optional[ProxyTypes] = None,
        client_dns_resolve: Optional[bool] = None,
        cache_storage: Optional[AsyncBaseStorage] = None,
        cache_folder_path: Optional[Union[str, Path]] = None,
        cache_sqlite_url: Optional[str] = None,
        cache_redis_url: Optional[str] = None,
        cache_ttl: Optional[int] = None,
        trust_env: bool = True,
    ):
        self._timeout = self._int_var("client_timeout", client_timeout, trust_env)
        self._retries = self._int_var("client_retries", client_retries, trust_env)
        if self._retries is None:
            self._retries = 10
        self._retry_on_blocked_ip = (
            self._bool_var(
                "client_retry_on_blocked_ip", client_retry_on_blocked_ip, trust_env
            )
            or False
        )
        self._proxy = self._str_var("client_proxy", client_proxy, trust_env)
        self._dns_resolve = (
            self._bool_var("client_dns_resolve", client_dns_resolve, trust_env) or False
        )

        self._cache_folder_path = self._str_var(
            "cache_folder_path", cache_folder_path, trust_env
        )
        self._cache_sqlite_url = self._str_var(
            "cache_sqlite_url", cache_sqlite_url, trust_env
        )
        self._cache_redis_url = self._str_var(
            "cache_redis_url", cache_redis_url, trust_env
        )
        self._cache_ttl = self._int_var("cache_ttl", cache_ttl, trust_env)
        self._cache_storage = cache_storage
        if (
            sum(
                [
                    int(bool(i))
                    for i in (
                        self._cache_folder_path,
                        self._cache_sqlite_url,
                        self._cache_redis_url,
                        self._cache_ttl,
                        self._cache_storage,
                    )
                ]
            )
            > 1
        ):
            raise ValueError("Допустимо выбрать только один вариант хранилища кэша")

        self._client: Optional[AsyncClient] = None
        self._last_response: Optional[Response] = None

    async def _build_cache_storage(self) -> Optional[AsyncBaseStorage]:
        if self._cache_folder_path is not None:
            return AsyncFileStorage(
                base_path=Path(self._cache_folder_path), ttl=self._cache_ttl
            )
        elif self._cache_sqlite_url is not None:
            if anysqlite is None:
                raise RuntimeError(
                    "Не установлены необходимые модули для кэширования этого типа. "
                    "Убедитесь, что вы установили `pynspd` с расширением `sqlite`.\n"
                    "```pip install pynspd[sqlite]```"
                )
            conn = await anysqlite.connect(
                self._cache_sqlite_url, check_same_thread=False
            )
            return AsyncSQLiteStorage(connection=conn, ttl=self._cache_ttl)
        elif self._cache_redis_url is not None:
            if redis is None:
                raise RuntimeError(
                    "Не установлены необходимые модули для кэширования этого типа. "
                    "Убедитесь, что вы установили `pynspd` с расширением `redis`.\n"
                    "```pip install pynspd[redis]```"
                )
            client = redis.Redis.from_url(self._cache_redis_url)
            return AsyncRedisStorage(client=client, ttl=self._cache_ttl)
        return None

    async def _build_client(self) -> AsyncClient:
        transport: AsyncBaseTransport = AsyncHTTPTransport(
            verify=SSL_CONTEXT, retries=self._retries, proxy=self._proxy
        )
        if self._cache_storage is None:
            self._cache_storage = await self._build_cache_storage()
        if self._cache_storage is not None:
            transport = AsyncCacheTransport(
                transport=transport,
                storage=self._cache_storage,
                controller=NSPD_CACHE_CONTROLLER,
            )

        host = self.DNS_HOST if not self._dns_resolve else self.IP_HOST
        return AsyncClient(
            base_url="https://" + host,
            timeout=self._timeout,
            headers={
                "User-Agent": ua_generator.generate().text,
                "Referer": self.DNS_URL,
                "Host": self.DNS_HOST,
            },
            transport=transport,
        )

    async def _rebuild_client_with_dns_resolve(self):
        assert not self._dns_resolve
        assert self._client is not None
        assert self._client.base_url == self.DNS_URL
        warnings.warn(
            "Текущий прокси не может сам разрешать доменные имена, "
            "рекомендуется использовать client_dns_resolve=True, "
            "чтобы клиент заранее подставлял IP-адреса и уменьшал нагрузку на соединение",
            stacklevel=7,
        )
        await self._client.aclose()
        self._dns_resolve = True
        self._client = await self._build_client()

    async def __aenter__(self):
        return self

    async def __aexit__(self, *exc):
        await self.close()

    async def close(self):
        """Завершение сессии"""
        if self._client is not None:
            await self._client.aclose()

    async def request(
        self,
        method: str,
        url: str,
        params: Optional[QueryParamTypes] = None,
        json: Optional[dict] = None,
    ) -> Response:
        """Базовый запрос к API НСПД"""
        logger.debug("Request %s", url)
        if self._client is None:
            self._client = await self._build_client()
        try:
            r = await self._client.request(method, url, params=params, json=json)
        except ProxyError as e:
            # Прокси не поддерживает DNS-resolve, необходимо использовать IP-адрес
            # https://github.com/encode/httpx/issues/203#issuecomment-1017726203
            msg = str(e)
            if (
                "Connection not allowed by ruleset" in msg
                or "503 Target host denied" in msg
            ) and not self._dns_resolve:
                logger.debug("Proxy can't resolve dns; change to ip mode")
                await self._rebuild_client_with_dns_resolve()
                return await self.request(method, url, params, json)
            raise e
        self._last_response = r
        code = r.status_code
        if code == 403:
            raise err.BlockedIP(r)
        if code == 404:
            raise err.NotFound(r)
        if code == 429:
            raise err.TooManyRequests(r)
        if code >= 500:
            raise err.PynspdServerError(r)
        if not str(code).startswith("2"):
            raise err.PynspdResponseError(r)
        return r

    @retry_on_http_error
    async def safe_request(
        self,
        method: str,
        url: str,
        params: Optional[QueryParamTypes] = None,
        json: Optional[dict] = None,
    ) -> Response:
        """Базовый запрос к api НСПД с обработкой ошибок"""
        return await self.request(method, url, params, json)

    ####################
    ### QUERY SEARCH ###
    ####################

    @retry_on_http_error
    async def _search(self, params: dict[str, Any]) -> Optional[list[NspdFeature]]:
        """Базовый поисковый запрос на НСПД"""
        try:
            r = await self.request(
                "get", "/api/geoportal/v2/search/geoportal", params=params
            )
            return SearchResponse.model_validate(r.json()).data.features
        except err.NotFound:
            return None

    async def search(
        self, query: str, theme_id: ThemeId = ThemeId.REAL_ESTATE_OBJECTS
    ) -> Optional[list[NspdFeature]]:
        """Поисковой запрос по предустановленной теме

        Args:
            query: Поисковой запрос
            theme_id:
                Вид объекта (кадастровое деление, объект недвижимости и т.д.).
                По умолчанию: объекты недвижимости

        Returns:
            Положительный ответ от сервиса, либо None, если ничего не найдено
        """
        return await self._search(
            params={
                "query": query,
                "thematicSearchId": theme_id.value,
            }
        )

    async def search_in_layer(
        self, query: str, layer_def: Type[Feat]
    ) -> Optional[list[Feat]]:
        """Поиск по определению слоя

        Args:
            query: Поисковой запрос
            layer_def: Определение слоя

        Returns:
            Массив валидированных объектов, если что-то найдено
        """
        raw_features = await self._search(
            params={
                "query": query,
                "layersId": layer_def.layer_meta.layer_id,
            }
        )
        return self._cast_features_to_layer_defs(raw_features, layer_def)

    async def search_in_layers(
        self, query: str, *layer_defs: Type[Feat]
    ) -> Optional[list[NspdFeature]]:
        """Поиск по определениям слоев

        Args:
            query: Поисковой запрос
            layer_defs: Определения слоев

        Returns:
            Массив невалидированных объектов, если что-то найдено
        """
        return await self._search(
            params={
                "query": query,
                "layersId": [i.layer_meta.layer_id for i in layer_defs],
            }
        )

    async def find(
        self, query: str, theme_id: ThemeId = ThemeId.REAL_ESTATE_OBJECTS
    ) -> Optional[NspdFeature]:
        """Найти объект по предустановленной теме

        Args:
            query: Поисковой запрос
            theme_id:
                Вид объекта (кадастровое деление, объект недвижимости и т.д.).
                По умолчанию: объекты недвижимости

        Returns:
            Положительный ответ от сервиса, либо None, если ничего не найдено
        """
        return self._filter_search_by_query(await self.search(query, theme_id), query)

    async def find_in_layer(self, query: str, layer_def: Type[Feat]) -> Optional[Feat]:
        """Найти объект по определению слоя

        Args:
            query: Поисковой запрос
            layer_def: Определение слоя

        Returns:
            Валидированная модель слоя, если найдено
        """
        return self._filter_search_by_query(
            await self.search_in_layer(query, layer_def), query
        )

    ######################
    ### POLYGON SEARCH ###
    ######################

    @retry_on_http_error
    async def _search_in_contour(
        self,
        countour: Union[Polygon, MultiPolygon],
        *category_ids: int,
    ) -> Optional[list[NspdFeature]]:
        """Поиск объектов в контуре по ID категорий слоев

        Args:
            countour: Геометрический объект с контуром
            category_ids: ID категорий слоев

        Returns:
            Список объектов, пересекающихся с контуром, если найден хоть один
        """
        feature_geojson = json.loads(to_geojson(countour))
        feature_geojson["crs"] = {
            "type": "name",
            "properties": {"name": "EPSG:4326"},
        }
        payload = {
            "categories": [{"id": id_} for id_ in category_ids],
            "geom": {
                "type": "FeatureCollection",
                "features": [
                    {"geometry": feature_geojson, "type": "Feature", "properties": {}}
                ],
            },
        }
        try:
            response = await self.request(
                "post",
                "/api/geoportal/v1/intersects",
                params={"typeIntersect": "fullObject"},
                json=payload,
            )
            return self._validate_feature_collection_response(response)
        except err.PynspdServerError as e:
            if '"code":400104' in e.response.text:
                raise err.TooBigContour from e
            raise e
        except json.decoder.JSONDecodeError as e:
            raise err.TooBigContour from e

    async def search_in_contour(
        self,
        countour: Union[Polygon, MultiPolygon],
        layer_def: Type[Feat],
    ) -> Optional[list[Feat]]:
        """Поиск объектов слоя в контуре

        Args:
            countour: Геометрический объект с контуром
            layer_def: Модель слоя

        Raises:
            TooBigContour: НСПД не может обработать такой контур

        Returns:
            Список объектов, пересекающихся с контуром, если найден хоть один
        """
        raw_features = await self._search_in_contour(
            countour, layer_def.layer_meta.category_id
        )
        return self._cast_features_to_layer_defs(raw_features, layer_def)

    async def _iter_search_in_box(
        self,
        xmin: float,
        ymin: float,
        xmax: float,
        ymax: float,
        layer_def: Type[Feat],
    ) -> AsyncGenerator[Feat, None]:
        """Рекурсивный поиск объектов в границах"""

        def split_extent(xmin: float, ymin: float, xmax: float, ymax: float):
            midx = (xmax + xmin) / 2
            midy = (ymax + ymin) / 2
            yield xmin, ymin, midx, midy
            yield midx, midy, xmax, ymax
            yield midx, ymin, xmax, midy
            yield xmin, midy, midx, ymax

        try:
            logger_prefix = "Search [%.2f, %.2f, %.2f, %.2f]: " % (
                xmin,
                ymin,
                xmax,
                ymax,
            )
            logger.debug(logger_prefix + "start")
            feats = await self.search_in_contour(box(xmin, ymin, xmax, ymax), layer_def)
            if feats is None:
                logger.debug(logger_prefix + "empty")
                return
            logger.debug(logger_prefix + "success")
            for f in feats:
                yield f
        except err.TooBigContour:
            logger.debug(logger_prefix + "failed, split tiles")
            for sp_xmin, sp_ymin, sp_xmax, sp_ymax in split_extent(
                xmin, ymin, xmax, ymax
            ):
                async for f in self._iter_search_in_box(
                    sp_xmin, sp_ymin, sp_xmax, sp_ymax, layer_def
                ):
                    yield f

    async def search_in_contour_iter(
        self,
        countour: Union[Polygon, MultiPolygon],
        layer_def: Type[Feat],
        *,
        only_intersects: bool = False,
    ) -> AsyncGenerator[Feat, None]:
        """Поиск объектов в указанных границах.

        Внимание: количество запросов кратно зависит от площади поиска.
        Если вы хотите вручную обрабатывать ошибку `TooBigContour`,
        используйте метод `search_in_contour(...)`

        Args:
            countour: Геометрический объект с контуром
            layer_def: Модель слоя
            only_intersects:
                Возвращать только те объекты,
                которые пересекаются с изначальным контуром. По умолчанию False

        Returns:
            Генератор объектов слоя в указанной области
        """
        cache = set()
        xmin, ymin, xmax, ymax = countour.bounds
        async for feat in self._iter_search_in_box(xmin, ymin, xmax, ymax, layer_def):
            cache_id = md5(feat.model_dump_json().encode()).hexdigest()
            if cache_id in cache:
                continue
            cache.add(cache_id)
            if only_intersects and not countour.intersects(feat.geometry.to_shape()):
                continue
            yield feat

    ####################
    ### POINT SEARCH ###
    ####################

    @retry_on_http_error
    async def _search_at_point(
        self, pt: Point, layer_id: int
    ) -> Optional[list[NspdFeature]]:
        """Поиск объектов слоя в точке"""
        tile_size = 512
        tile = mercantile.tile(
            pt.x, pt.y, zoom=24
        )  # zoom=24 должно быть достаточно для самого точного совпадения
        tile_bounds = mercantile.bounds(tile)
        i = np.interp(pt.x, [tile_bounds.west, tile_bounds.east], [0, tile_size])
        j = np.interp(pt.y, [tile_bounds.south, tile_bounds.north], [0, tile_size])
        bbox = ",".join(map(str, tile_bounds))
        params = {
            "REQUEST": "GetFeatureInfo",
            "SERVICE": "WMS",
            "VERSION": "1.3.0",
            "INFO_FORMAT": "application/json",
            "FORMAT": "image/png",
            "STYLES": "",
            "TRANSPARENT": "true",
            "QUERY_LAYERS": layer_id,
            "LAYERS": layer_id,
            "WIDTH": tile_size,
            "HEIGHT": tile_size,
            "I": int(i),
            "J": tile_size - int(j),  # отсчет координат для пикселей ведется сверху
            "CRS": "EPSG:4326",  # CRS для bbox
            "BBOX": bbox,
            "FEATURE_COUNT": "10",  # Иначе вернет только один, даже если попало на границу
        }
        response = await self.request(
            "get",
            f"/api/aeggis/v3/{layer_id}/wms",
            params=params,  # type: ignore[arg-type]
        )
        return self._validate_feature_collection_response(response)

    async def search_at_point(
        self, pt: Point, layer_def: Type[Feat]
    ) -> Optional[list[Feat]]:
        """Поиск объектов слоя в точке (с типизацией)

        Args:
            pt: Точка поиска
            layer_def: Тип слоя

        Returns:
            Типизированный список объектов, если найдены
        """
        raw_features = await self._search_at_point(pt, layer_def.layer_meta.layer_id)
        return self._cast_features_to_layer_defs(raw_features, layer_def)

    async def search_at_coords(
        self, lat: float, lng: float, layer_def: Type[Feat]
    ) -> Optional[list[Feat]]:
        """Поиск объектов слоя в координатах

        Args:
            lat: Широта
            lng: Долгота
            layer_def: Тип слоя

        Returns:
            Типизированный список объектов, если найдены
        """
        return await self.search_at_point(Point(lng, lat), layer_def)

    ####################
    ### TAB REQUESTS ###
    ####################

    @retry_on_http_error
    async def _tab_request(
        self, feat: NspdFeature, tab_class: str, type_: Literal["values", "group"]
    ) -> Optional[dict]:
        if feat.properties.options.no_coords:
            params = {
                "tabClass": tab_class,
                "objdocId": feat.properties.options.objdoc_id,
                "registersId": feat.properties.options.registers_id,
            }
        else:
            params = {
                "tabClass": tab_class,
                "categoryId": feat.properties.category,
                "geomId": feat.id,
            }
        try:
            r = await self.request(
                "get", f"/api/geoportal/v1/tab-{type_}-data", params=params
            )
            return r.json()
        except err.NotFound:
            return None

    async def _tab_values_request(
        self, feat: NspdFeature, tab_class: str
    ) -> Optional[list[str]]:
        resp = await self._tab_request(feat, tab_class, "values")
        if resp is None:
            return None
        return NspdTabResponse.model_validate(resp).value

    async def _tab_groups_request(
        self, feat: NspdFeature, tab_class: str
    ) -> Optional[dict[str, Optional[list[str]]]]:
        resp = await self._tab_request(feat, tab_class, "group")
        if resp is None:
            return None
        item = NspdTabGroupResponse.model_validate(resp).object
        data = {}
        for i in item:
            title = re.sub(r"[\s:]+$", "", i.title)
            data[title] = i.value
        if len(data) == 0:
            return None
        return data

    async def tab_land_parts(self, feat: NspdFeature) -> Optional[list[str]]:
        """Получение данных с вкладки \"Части ЗУ\" """
        return await self._tab_values_request(feat, "landParts")

    async def tab_land_links(self, feat: NspdFeature) -> Optional[list[str]]:
        """Получение данных с вкладки \"Связанные ЗУ\" """
        return await self._tab_values_request(feat, "landLinks")

    async def tab_permission_type(self, feat: NspdFeature) -> Optional[list[str]]:
        """Получение данных с вкладки \"Виды разрешенного использования\" """
        return await self._tab_values_request(feat, "permissionType")

    async def tab_composition_land(self, feat: NspdFeature) -> Optional[list[str]]:
        """Получение данных с вкладки \"Состав ЕЗП\" """
        return await self._tab_values_request(feat, "compositionLand")

    async def tab_build_parts(self, feat: NspdFeature) -> Optional[list[str]]:
        """Получение данных с вкладки \"Части ОКС\" """
        return await self._tab_values_request(feat, "buildParts")

    async def tab_objects_list(
        self, feat: NspdFeature
    ) -> Optional[dict[str, Optional[list[str]]]]:
        """
        Получение данных с вкладки \"Объекты\"
        """
        return await self._tab_groups_request(feat, "objectsList")

    async def get_tab_data(self, feat: NspdFeature, tab_name: TabTitle):
        """Получение данных с указанной вкладки"""
        match tab_name:
            case "Части ЗУ":
                return await self.tab_land_parts(feat)
            case "Связанные ЗУ":
                return await self.tab_land_links(feat)
            case "Виды разрешенного использования":
                return await self.tab_permission_type(feat)
            case "Состав ЕЗП":
                return await self.tab_composition_land(feat)
            case "Части ОКС":
                return await self.tab_build_parts(feat)
            case "Объекты":
                return await self.tab_objects_list(feat)

    #################
    ### SHORTCUTS ###
    #################

    async def find_landplot(self, query: str) -> Optional[Layer36048Feature]:
        """Найти ЗУ по кадастровому номеру"""
        return self._filter_search_by_query(await self.search_landplots(query), query)

    async def find_building(self, query: str) -> Optional[Layer36049Feature]:
        """Найти ОКС по кадастровому номеру"""
        return self._filter_search_by_query(await self.search_buildings(query), query)

    async def search_landplots_at_point(
        self, pt: Point
    ) -> Optional[list[Layer36048Feature]]:
        """Поиск ЗУ в точке"""
        return await self.search_at_point(pt, Layer36048Feature)

    async def search_buildings_at_point(
        self, pt: Point
    ) -> Optional[list[Layer36049Feature]]:
        """Поиск ОКС в точке"""
        return await self.search_at_point(pt, Layer36049Feature)

    async def search_landplots(self, cn: str) -> Optional[list[Layer36048Feature]]:
        """Поиск ЗУ по кадастровому номеру"""
        return await self.search_in_layer(cn, Layer36048Feature)

    async def search_buildings(self, cn: str) -> Optional[list[Layer36049Feature]]:
        """Поиск ОКС по кадастровому номеру"""
        return await self.search_in_layer(cn, Layer36049Feature)

    async def search_landplots_at_coords(
        self, lat: float, lng: float
    ) -> Optional[list[Layer36048Feature]]:
        """Поиск ЗУ в координатах"""
        return await self.search_at_coords(lat, lng, Layer36048Feature)

    async def search_buildings_at_coords(
        self, lat: float, lng: float
    ) -> Optional[list[Layer36049Feature]]:
        """Поиск ОКС в координатах"""
        return await self.search_at_coords(lat, lng, Layer36049Feature)

    async def search_landplots_in_contour(
        self, countour: Union[Polygon, MultiPolygon]
    ) -> Optional[list[Layer36048Feature]]:
        """Поиск ЗУ в контуре"""
        return await self.search_in_contour(countour, Layer36048Feature)

    async def search_buildings_in_contour(
        self, countour: Union[Polygon, MultiPolygon]
    ) -> Optional[list[Layer36049Feature]]:
        """Поиск ОКС в контуре"""
        return await self.search_in_contour(countour, Layer36049Feature)



import pytest
from shapely import wkt
from shapely.geometry import (
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Polygon,
    box,
)

from pynspd import AsyncNspd, NspdFeature
from pynspd.errors import TooBigContour
from pynspd.schemas import Layer36048Feature, Layer36049Feature, Layer37578Feature


@pytest.mark.asyncio(scope="session")
async def test_find(async_api: AsyncNspd):
    feat = await async_api.find("77:02:0021001:5304")
    assert feat is not None
    assert feat.properties.options.model_dump()["type"] == "Машино-место"


@pytest.mark.asyncio(scope="session")
async def test_find_non_exists(async_api: AsyncNspd):
    feat = await async_api.find("77:02:0021001:5304111111")
    assert feat is None


@pytest.mark.asyncio(scope="session")
async def test_find_in_layer(async_api: AsyncNspd):
    lf_def = NspdFeature.by_title("ЗОУИТ объектов энергетики, связи, транспорта")
    assert lf_def == Layer37578Feature
    feat = await async_api.find_in_layer("Останкинская телебашня", lf_def)
    assert isinstance(feat, lf_def)
    assert len(feat.properties.options.model_dump_human_readable()) > 0


@pytest.mark.asyncio(scope="session")
async def test_find_in_layer_non_exists(async_api: AsyncNspd):
    feat = await async_api.find_in_layer("77:02:0021001:5304111111", Layer36048Feature)
    assert feat is None


@pytest.mark.asyncio(scope="session")
async def test_search_landplots(async_api: AsyncNspd):
    feat = await async_api.find_landplot("77:05:0001005:19")
    assert feat is not None
    assert feat.properties.options.land_record_type == "Земельный участок"
    assert isinstance(feat.geometry.to_shape(), Polygon)
    assert isinstance(feat.geometry.to_multi_shape(), MultiPolygon)


@pytest.mark.asyncio(scope="session")
async def test_search_buildings(async_api: AsyncNspd):
    feat = await async_api.find_building("77:03:0001001:3030")
    assert feat is not None
    assert feat.properties.options.build_record_type_value == "Здание"
    assert isinstance(feat.geometry.to_shape(), Polygon)


@pytest.mark.asyncio(scope="session")
async def test_search_landplots_in_contour(async_api: AsyncNspd):
    contour = wkt.loads(
        "Polygon ((37.62381 55.75345, 37.62577 55.75390, 37.62448 55.75278, 37.62381 55.75345))"
    )
    features = await async_api.search_landplots_in_contour(contour)
    assert features is not None
    assert all([isinstance(i, Layer36048Feature) for i in features])
    cns = [i.properties.options.cad_num for i in features]
    assert set(["77:01:0001011:8", "77:01:0001011:14", "77:01:0001011:16"]) == set(cns)


@pytest.mark.asyncio(scope="session")
async def test_search_buildings_in_contour(async_api: AsyncNspd):
    contour = wkt.loads(
        "Polygon ((37.62381 55.75345, 37.62577 55.75390, 37.62448 55.75278, 37.62381 55.75345))"
    )
    features = await async_api.search_buildings_in_contour(contour)
    assert features is not None
    assert all([isinstance(i, Layer36049Feature) for i in features])
    cns = [i.properties.options.cad_num for i in features]
    assert set(["77:01:0001011:1164", "77:01:0001011:1002"]) == set(cns)


@pytest.mark.asyncio(scope="session")
async def test_search_in_contour_empty(async_api: AsyncNspd):
    contour = wkt.loads(
        "Polygon ((37.63215 55.75588, 37.63214 55.75557, 37.63271 55.75570, 37.63215 55.75588))"
    )
    features = await async_api.search_landplots_in_contour(contour)
    assert features is None


@pytest.mark.asyncio(scope="session")
async def test_search_landplots_at_coords(async_api: AsyncNspd):
    features = await async_api.search_landplots_at_coords(55.78729561, 37.54658156)
    assert features is None
    features = await async_api.search_landplots_at_coords(55.787139958, 37.546440653)
    assert features is not None and len(features) == 1
    assert features[0].properties.options.cad_num == "77:09:0005008:11446"


@pytest.mark.asyncio(scope="session")
async def test_search_buildings_at_point(async_api: AsyncNspd):
    features = await async_api.search_buildings_at_coords(55.786436698, 37.547790951)
    assert features is None
    features = await async_api.search_buildings_at_coords(55.786436698, 37.547785813)
    assert features is not None and len(features) == 1
    assert features[0].properties.options.cad_num == "77:09:0005014:1044"


@pytest.mark.asyncio(scope="session")
async def test_search_wrong_result(async_api: AsyncNspd):
    features = await async_api.search("77:1:3033:1031")
    assert features is not None
    feat = await async_api.find("77:1:3033:1031")
    assert feat is None


@pytest.mark.asyncio(scope="session")
async def test_search_layers(async_api: AsyncNspd):
    features = await async_api.search_in_layers(
        "Обнинск",
        NspdFeature.by_title("Муниципальные образования (полигональный)"),
        NspdFeature.by_title("Населённые пункты (полигоны)"),
    )
    assert features is not None
    assert len(features) == 2


@pytest.mark.asyncio(scope="session")
async def test_search_in_big_contour(async_api: AsyncNspd):
    contour = box(36.80, 55.14, 37.39, 55.58)
    layer_def = NspdFeature.by_title("Земельные участки из ЕГРН")
    with pytest.raises(TooBigContour):
        async_api._retries = 0
        await async_api.search_in_contour(contour, layer_def)

    async for feat in async_api.search_in_contour_iter(contour, layer_def):
        assert feat is not None
        break


@pytest.mark.asyncio(scope="session")
async def test_search_in_linear_layer(async_api: AsyncNspd):
    feat = await async_api.find_in_layer(
        "60:09-3.7", NspdFeature.by_title("Муниципальные образования (линейный)")
    )
    assert feat is not None
    assert isinstance(feat.geometry.to_shape(), LineString)
    assert isinstance(feat.geometry.to_multi_shape(), MultiLineString)


@pytest.mark.asyncio(scope="session")
async def test_search_in_point_layer(async_api: AsyncNspd):
    feats = await async_api.search_in_contour(
        box(
            34.93710,
            61.90981,
            35.43174,
            62.15627,
        ),
        NspdFeature.by_title("Объекты туристского интереса"),
    )
    assert feats is not None
    feat = feats[0]
    # объекта в точечном слое с типом Point не нашел
    # assert isinstance(feat.geometry.to_shape(), LineString)
    assert isinstance(feat.geometry.to_multi_shape(), MultiPoint)



from enum import Enum
from typing import Literal


class ThemeId(Enum):
    """Вид объекта (для поиска по категории)"""

    REAL_ESTATE_OBJECTS = 1
    """Объекты недвижимости"""

    CADASTRAL_DIVISION = 2
    """Кадастровое деление"""

    ADMINISTRATIVE_TERRITORIAL_DIVISION = 4
    """Административно-территориальное деление"""

    ZONES_AND_TERRITORIES = 5
    """Зоны и территории"""

    TERRITORIAL_ZONES = 7
    """Территориальные зоны"""

    COMPLEXES_OF_OBJECTS = 15
    """Комплексы объектов"""


TabTitle = Literal[
    "Части ЗУ",
    "Связанные ЗУ",
    "Виды разрешенного использования",
    "Состав ЕЗП",
    "Части ОКС",
    "Объекты",
]



import httpx
import pytest
from pytest_httpx import HTTPXMock

from pynspd import Nspd
from pynspd.errors import BlockedIP, PynspdResponseError, PynspdServerError


def test_429_error(httpx_mock: HTTPXMock, api: Nspd):
    httpx_mock.add_response(status_code=429)
    httpx_mock.add_response(status_code=200)
    r = api.safe_request("get", "/api")
    assert r.status_code == 200
    assert len(httpx_mock.get_requests()) == 2


def test_400_error(httpx_mock: HTTPXMock, api: Nspd):
    httpx_mock.add_response(status_code=400)
    with pytest.raises(PynspdResponseError) as e:
        api.safe_request("get", "/api")
    assert e.value.response.status_code == 400


def test_500_error(httpx_mock: HTTPXMock, api: Nspd):
    api._retries = 1
    httpx_mock.add_response(status_code=500)
    httpx_mock.add_response(status_code=500)
    with pytest.raises(PynspdServerError) as e:
        api.safe_request("get", "/api")
    assert e.value.response.status_code == 500
    assert len(httpx_mock.get_requests()) > api._retries


def test_remote_disconnect_error(httpx_mock: HTTPXMock, api: Nspd):
    httpx_mock.add_exception(httpx.RemoteProtocolError("Unexpected disconnect"))
    httpx_mock.add_response(status_code=200)
    r = api.safe_request("get", "/api")
    assert r.status_code == 200


def test_handled_403_error(httpx_mock: HTTPXMock, api: Nspd):
    api._retry_on_blocked_ip = True
    httpx_mock.add_response(status_code=403)
    httpx_mock.add_response()
    api.safe_request("get", "/api")
    assert len(httpx_mock.get_requests()) == 2


def test_unhandled_403_error(httpx_mock: HTTPXMock, api: Nspd):
    api._retry_on_blocked_ip = False
    httpx_mock.add_response(status_code=403)
    with pytest.raises(BlockedIP):
        api.safe_request("get", "/api")



import pytest

from pynspd import AsyncNspd, NspdFeature


async def get_feat(async_api: AsyncNspd, cn: str) -> NspdFeature:
    feat = await async_api.find(cn)
    assert feat is not None
    return feat


@pytest.mark.asyncio(scope="session")
async def test_tab_land_parts(async_api: AsyncNspd):
    feat = await get_feat(async_api, "50:27:0000000:134535")
    res = await async_api.get_tab_data(feat, "Части ЗУ")
    assert res is not None
    assert len(res) > 10


@pytest.mark.asyncio(scope="session")
async def test_tab_land_links(async_api: AsyncNspd):
    feat = await get_feat(async_api, "77:01:0001001:1024")
    res = await async_api.get_tab_data(feat, "Связанные ЗУ")
    assert res == ["77:01:0001001:1001", "77:01:0001001:1514", "77:01:0001001:1829"]


@pytest.mark.asyncio(scope="session")
async def test_tab_permission_type(async_api: AsyncNspd):
    feat = await get_feat(async_api, "50:27:0000000:134535")
    res = await async_api.get_tab_data(feat, "Виды разрешенного использования")
    assert res is not None
    assert len(res) == 1
    assert res[0].startswith("заготовка древесины;")


@pytest.mark.asyncio(scope="session")
async def test_tab_build_parts(async_api: AsyncNspd):
    feat = await get_feat(async_api, "77:01:0001001:1024")
    res = await async_api.get_tab_data(feat, "Части ОКС")
    assert res == ["77:01:0001001:1024/1"]


@pytest.mark.asyncio(scope="session")
async def test_tab_objects_list(async_api: AsyncNspd):
    feat = await get_feat(async_api, "77:01:0001001:1024")
    res = await async_api.get_tab_data(feat, "Объекты")
    assert res is not None
    assert len(res) == 4
    assert res["Помещения (количество)"] == ["26"]


@pytest.mark.asyncio(scope="session")
async def test_tab_composition_land(async_api: AsyncNspd):
    feat = await get_feat(async_api, "48:06:0000000:111")
    assert feat.properties.options.no_coords
    res = await async_api.get_tab_data(feat, "Состав ЕЗП")
    assert res == [
        "48:06:1620101:209",
        "48:06:1620101:210",
        "48:06:1630501:28",
        "48:06:1630501:29",
    ]


@pytest.mark.asyncio(scope="session")
async def test_unsuitable_feat(async_api: AsyncNspd):
    feat = await get_feat(async_api, "50:27:0000000:134535")
    res = await async_api.tab_build_parts(feat)
    assert res is None


@pytest.mark.asyncio(scope="session")
async def test_empty_tab(async_api: AsyncNspd):
    feat = await get_feat(async_api, "50:27:0000000:134535")
    res = await async_api.tab_composition_land(feat)
    assert res is None


@pytest.mark.asyncio(scope="session")
async def test_tab_objects_key_format(async_api: AsyncNspd):
    feat = await get_feat(async_api, "50:21:0110501:792")
    res = await async_api.tab_objects_list(feat)
    assert res is not None
    assert "Объект недвижимости" in res



# autogenerated
from typing import Literal

LayerTitle = Literal[
    "Водная эрозия",
    "Линейная эрозия",
    "Ветровая эрозия",
    "Опустынивание",
    "Переувлажнение",
    "Подтопление",
    "Заболачивание",
    "Затопление",
    "Захламление",
    "Обвально-осыпные и оползневые процессы",
    "Абразия",
    "Кадастровые округа",
    "Территории объектов культурного наследия",
    "Иные нарушенные земли ",
    "Кадастровые районы ",
    "Нарушенные земли при наземном строительстве ",
    "Кадастровые кварталы",
    "Нарушенные земли при гидротехническом строительстве",
    "Нарушенные земли при недропользовании",
    "Нарушенные земли при промышленном лесопользовании",
    "Нарушенные земли при сельскохозяйственном освоении",
    "Нарушенные земли при проведении геологоразведочных, испытательных, эксплуатационных и иных работ",
    "Нарушенные земли при складировании и захоронении промышленных отходов, загрязнение земель",
    "Государственная граница Российской Федерации",
    "Гари",
    "Негативный процесс отсутствует",
    "Субъекты Российской Федерации (линии)",
    "Субъекты Российской Федерации (полигоны)",
    "Засоление",
    "Муниципальные образования (полигональный)",
    "Муниципальные образования (линейный)",
    "Населённые пункты (полигоны)",
    "Населённые пункты (линии)",
    "Земельные участки из ЕГРН",
    "Земельные участки, образуемые по схеме расположения земельного участка",
    "Земельные участки, выставленные на аукцион ",
    "Земельные участки, свободные от прав третьих лиц",
    "Земельные участки, образуемые по проекту межевания территории",
    "Здания",
    "Сооружения",
    "Объекты незавершенного строительства",
    "Единые недвижимые комплексы",
    "Предприятие как имущественный комплекс",
    "ЗОУИТ объектов культурного наследия",
    "ЗОУИТ объектов энергетики, связи, транспорта",
    "ЗОУИТ природных территорий",
    "ЗОУИТ охраняемых объектов и безопасности",
    "Иные ЗОУИТ",
    "Территориальные зоны",
    "Красные линии ",
    "Особо охраняемые природные территории ",
    "Охотничьи угодья ",
    "Лесничества",
    "Граница лесопарка",
    "Береговые линии (границы водных объектов) (полигональный)",
    "Береговые линии (границы водных объектов)(линейный)",
    "Особые экономические зоны",
    "Территории опережающего развития",
    "Игорные зоны",
    "Территории выполнения комплексных кадастровых работ",
    "Территория проведения мероприятий по ликвидации накопленного вреда окружающей среде, образовавшегося в результате производства химической продукции в г. Усолье-Сибирское Иркутской области",
    "Кадастровая стоимость объекта",
    "Удельный показатель кадастровой стоимости",
]



from typing import ClassVar, Generic, TypeVar

from geojson_pydantic import Feature

from pynspd.schemas.geometries import Geometry
from pynspd.schemas.layer_configs import LayerNode
from pynspd.schemas.properties import NspdProperties

Props = TypeVar("Props", bound="NspdProperties")
Geom = TypeVar("Geom", bound="Geometry")


class BaseFeature(Feature[Geom, Props], Generic[Geom, Props]):
    layer_meta: ClassVar[LayerNode]
    """Определение слоя из конфига НСПД"""

    # переопределяем поля из geojson-pydantic, т.к. там они не обязательные
    geometry: Geom
    properties: Props

    def __repr__(self):
        title = self.properties.get_title()
        if title is not None:
            return f"NspdFeature<{self.properties.category_name}: {title}>"
        return f"NspdFeature<{self.properties.category_name}>"



import os

import pytest

from pynspd import Nspd


def test_env_warning():
    os.environ["PYNSPD_CLIENT_TIMEOUT"] = "10"
    with pytest.warns():
        api = Nspd(client_timeout=5)
        assert api._timeout == 10



import asyncio
import re
import warnings
from functools import partial
from typing import Optional

from jinja2 import Template

from pynspd import AsyncNspd
from pynspd._async.api import retry_on_http_error
from pynspd.schemas import _autogen_features as cached_feature_configs
from pynspd.schemas.base_feature import BaseFeature
from pynspd.schemas.layer_configs import Card, CardField, LayerNode, LayersTree

CLIENT = AsyncNspd()
request = partial(retry_on_http_error(AsyncNspd.request), CLIENT)


TYPES_TEMPLATE = """# autogenerated
from typing import Literal


LayerTitle = Literal[
    {%- for layer in layers %}
    "{{ layer.title }}",
    {%- endfor %}
]
"""

SCHEMAS_TEMPLATE = """# autogenerated
from datetime import datetime
from typing import Union, Optional

from pynspd.schemas.base_feature import BaseFeature
from pynspd.schemas.geometries import Point, MultiPoint, LineString, MultiLineString, Polygon, MultiPolygon
from pynspd.schemas.layer_configs import LayerNode
from pynspd.schemas.properties import NspdProperties, OptionProperties


{% for category_id, (title_key, fields) in layers_fields.items() %}
class Options{{ category_id }}(OptionProperties): {% if fields|length == 0 %}...
    {% else %}
    {% if title_key %}title_key = "{{ title_key }}"{% endif %}
    {%- for field in fields %}
    {{ field.key_value }}: Optional[{{field.key_type}}] = None
    \"\"\"{{field.key_name}}{% if field.postfix != None %} ({{ field.postfix }}){% endif %}\"\"\"
    {%- endfor %}
    {% endif %}
{% endfor %}


{% for layer in layers %}
class Layer{{ layer.layer_id }}Feature(BaseFeature[
    Union[Multi{{ layer.geometry_type }}, {{ layer.geometry_type }}{%- if layer.geometry_type == 'Polygon' %}, Point{% endif -%}],
    NspdProperties[Options{{ layer.category_id }}]
]):
    \"\"\"{{ layer.title }}\"\"\"
    layer_meta = LayerNode.model_validate({{ layer.model_dump(by_alias=True) }})
{% endfor %}
"""

OVERLOAD_TEMPLATE = """{%- for layer in layers %}
    @overload
    @classmethod
    def by_title(cls, title: Literal["{{ layer.title }}"]) -> Type[auto.Layer{{ layer.layer_id }}Feature]: ...
{% endfor %}
    @overload
    @classmethod
    def by_title(cls, title: Any) -> NoReturn: ...
"""


def generate_files(layers: list[LayerNode], layers_fields: dict[int, Card]):
    output = Template(SCHEMAS_TEMPLATE).render(
        layers=layers, layers_fields=layers_fields
    )
    with open("src/pynspd/schemas/_autogen_features.py", "w", encoding="utf-8") as file:
        file.write(output)

    output = Template(TYPES_TEMPLATE).render(layers=layers)
    with open("src/pynspd/map_types/_autogen_layers.py", "w", encoding="utf-8") as file:
        file.write(output)

    output = Template(OVERLOAD_TEMPLATE).render(layers=layers)
    lines = []
    with open("src/pynspd/schemas/feature.py", "r", encoding="utf-8") as from_file:
        for line in from_file:
            lines.append(line)
    with open("src/pynspd/schemas/feature.py", "w", encoding="utf-8") as to_file:
        write_flag = True
        for line in lines:
            if "# END_AUTOGEN" in line:
                write_flag = True
            if write_flag:
                to_file.write(line)
            if "# START_AUTOGEN" in line:
                write_flag = False
                to_file.write(output)
                to_file.write("\n\n")


def get_cached_geometry_types() -> dict[int, str]:
    data = {}
    for i in dir(cached_feature_configs):
        if not re.match(r"Layer\d+Feature", i):
            continue
        cl: BaseFeature = getattr(cached_feature_configs, i)
        data[cl.layer_meta.layer_id] = LayerNode._geometry_type_serializer(
            cl.layer_meta.geometry_type
        )
    return data


async def get_layer_tree() -> LayersTree:
    r = await request(
        "get", "/api/geoportal/v1/layers-theme-tree", params={"themeId": 1}
    )
    data = r.json()
    # 01.04.2025 - исчезла geometryType, берем из предыдущих значений
    cached_geometry_types = get_cached_geometry_types()
    for layer in data["layers"]:
        if "geometryType" in layer:
            continue
        warnings.warn(
            "Not found geometry type in tree; cached used instead", stacklevel=2
        )
        layer["geometryType"] = cached_geometry_types.get(layer["layerId"], "POLYGON")
    tree = LayersTree.model_validate(data)
    return tree


async def get_category_card(
    category_id: int,
) -> tuple[int, tuple[Optional[str], list[CardField]]]:
    r = await request(
        "get", f"/api/geoportal/v1/geom-card-display-settings/{category_id}"
    )
    card = Card.model_validate(r.json())
    title_key = card.title.key_value

    fields = []
    for field in card.card:
        if field.key_value == "-":
            continue
        if field.key_value == "":
            continue
        if not field.key_value.isascii():
            continue
        if " " in field.key_value:
            continue
        fields.append(field)
    return category_id, (title_key, fields)


async def get_layers_fields(layers: list[LayerNode]) -> dict[int, list[CardField]]:
    layers_fields = {}
    categories = list(set([i.category_id for i in layers]))
    result = await asyncio.gather(*[get_category_card(i) for i in categories])
    layers_fields = dict(result)
    return layers_fields


async def main():
    tree = await get_layer_tree()
    layers = tree.layers
    layers_fields = await get_layers_fields(layers)
    generate_files(layers, layers_fields)


if __name__ == "__main__":
    asyncio.run(main())
